%% ----------------- Information --------------
% Implementation of Local Frequency Descriptor (LFD) Magnitude
% Please cite the following paper if you use the code.
% Reference: R. Maani, S. Kalra, and Y.H. Yang, "Noise robust rotation invariant features for texture classification", Pattern Recognition 46(8), pp. 21032116, 2013.
% 
% Copyright (C) 2013 Rouzbeh Maani  
% All rights reserved.
% 
% This program is free software; you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation; either version 2 of the License, or
% at your option any later version.
% See the GNU General Public License for more details.


%% -------------- External Codes ---------------
% The matlab files used from the implementation of clpb (http://www.comp.polyu.edu.hk/~cslzhang/code/CLBP.rar)
% ClassifyOnNN, ReadOutexTxt
% The matlab files used from the Oxford texture classification filter bank 
% DGFilters


%% ---------------------------------
%  This program computes the magnitude of local frequency descriptors (LFD)

%% ---------------------------------

function result = lfd_mag(I,R,N,FreqLevNo,all_radii)
% I: the image
% R: radius 
% N: number of samples
% FreqLevNo: number of frequency levels
error(nargchk(1,5,nargin));
if(nargin<4)
    FreqLevNo = N/8+1;
end
% ------------------------------- RADII -----------------------------------
if(nargin<5)
    radii = 18:3:45;
    radii2 = [10,radii];
else
    radii = all_radii(2:end);
    radii2 = all_radii(1:end-1);
end
%------------------------------ Local Functions ---------------------------
I=double(I);
[ysize xsize] = size(I);
if nargin==1
    pointsRelativeLocation=[-1 -1; -1 0; -1 1; 0 -1; -0 1; 1 -1; 1 0; 1 1];
    N = 8;
else
    pointsRelativeLocation=zeros(N,2);
    angle = 2*pi/N;
    for i = 1:N
        pointsRelativeLocation(i,1) = -R*sin((i-1)*angle);
        pointsRelativeLocation(i,2) = R*cos((i-1)*angle);
    end
end
miny = min(pointsRelativeLocation(:,1));
maxy = max(pointsRelativeLocation(:,1));
minx = min(pointsRelativeLocation(:,2));
maxx = max(pointsRelativeLocation(:,2));
bsizey = ceil(max(maxy,0))-floor(min(miny,0))+1;
bsizex = ceil(max(maxx,0))-floor(min(minx,0))+1;
origy = 1-floor(min(miny,0));
origx = 1-floor(min(minx,0));
if(xsize < bsizex || ysize < bsizey)
    error('Too small input I. Should be at least (2*R+1) x (2*R+1)');
end
dx = xsize-bsizex;
dy = ysize-bsizey;
lbpFunction = zeros(N,dy+1,dx+1);
for i = 1:N
    y = pointsRelativeLocation(i,1)+origy;
    x = pointsRelativeLocation(i,2)+origx;
    fy = floor(y); cy = ceil(y); ry = round(y);
    fx = floor(x); cx = ceil(x); rx = round(x);
    if (abs(x - rx) < 1e-6) && (abs(y - ry) < 1e-6)
        D = I(ry:ry+dy,rx:rx+dx);
    else
        ty = y-fy;
        tx = x-fx;
        w1 = (1-tx)*(1-ty);
        w2 = tx*(1-ty);
        w3 = (1-tx)*ty;
        w4 = tx*ty;
        D = w1*I(fy:fy+dy,fx:fx+dx)+w2*I(fy:fy+dy,cx:cx+dx)+w3*I(cy:cy+dy,fx:fx+dx)+w4*I(cy:cy+dy,cx:cx+dx);
    end  
    lbpFunction(i,:,:) = D;
end
%------------------------ Frequency Spectrum of lfd -----------------------
lbpFunctionFFT = fft(lbpFunction,[],1);
FerqChannels = abs(lbpFunctionFFT(1:FreqLevNo,:,:));
ctr = [size(FerqChannels,2) size(FerqChannels,3)]/2;
X = -ctr(2)+1:size(FerqChannels,3)-ctr(2);
Y = -ctr(1)+1:size(FerqChannels,2)-ctr(1);
[X,Y] = meshgrid(X,Y);
dist = sqrt(X.^2+Y.^2);
%------------------------------- FLITERS ----------------------------------
dirs = 21;
filter = DGFilters(size(X),dirs);
for i=1:size(filter,3)
    disk = dist<=25&dist>10;
    filter(:,:,i) = filter(:,:,i).*disk;
end
%----------------- Applying Filters to Frequency Channels -----------------
lfd_circular_norm1 = zeros(size(radii,2),FreqLevNo);
lfd_circular_norm2 = zeros(size(radii,2),FreqLevNo);
lfd_directional = zeros(size(filter,3),FreqLevNo);
for i=1:FreqLevNo
    lfd_f = abs(fftshift(fft2(squeeze(FerqChannels(i,:,:)))));
    for j=1:size(filter,3)
        lfd_directional(j,i) = sum(sum(lfd_f.*filter(:,:,j)));
    end
    lfd_directional(:,i) = lfd_directional(:,i)/sum(lfd_directional(:,i));
    for j=1:size(radii,2)
        disk = dist<=radii(j)&dist>radii2(j);
        lfd_f_masked = lfd_f.*disk;
        lfd_circular_norm2(j,i) = mean(lfd_f_masked(lfd_f_masked>0));
    end
    lfd_circular_norm1(:,i) = lfd_circular_norm2(:,i)/max(lfd_f(:));
end
lfd_circular_norm2 = lfd_circular_norm2./repmat(sum(lfd_circular_norm2,1),size(lfd_circular_norm2,1),1);
lfd_directional = [min(lfd_directional,[],1);median(lfd_directional,1);max(lfd_directional,[],1)];
%---------------------------- All Features --------------------------------
%result = [lfd_circular_norm2(:)',lfd_circular_norm1(:)',lfd_directional(:)'];
result = [lfd_circular_norm1(:)',lfd_circular_norm2(:)',lfd_directional(:)'];
result = result(:)';
end